@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 '''isort:skip_file'''
 __title__ = 'cpenv'
-__version__ = '0.5.1'
+__version__ = '0.5.2'
 __author__ = 'Dan Bradham'
 __email__ = 'danielbradham@gmail.com'
 __url__ = 'http://github.com/cpenv/cpenv'
@@ -12,9 +12,10 @@ __description__ = 'Cross-platform Python environment management.'
 # Local imports
 from .api import *
 from .repos import *
+from .reporter import *
 from .resolver import *
-from .vendor import yaml
 from .versions import *
+from . import vendor
 
 
 # Initialize cpenv
